//
// d o s l i b . c
//
// Small DOS library functions
//
// Jul. 8, 2003 by Wataru Nishida (http://www.wnishida.com)
// Sep.10, 2003 -- atoi was added
// Nov.25, 2003 -- putchar was replaced with dos_putchar
//		   __asm__ codes were rewritten with EXTENDED inline assembler
//

#include "realmode.h"

// Simple putchar function for DOS

static
void dos_putc(int ch) {
  Asm(	" movb	$2, %%ah	\n\t"	// Write to standard output device
	" int	$0x21		"	// Execute DOS function call
	:				// No output
	: "d" (ch)			// Input register (EDX = ch)
   );
 }

void dos_putchar(int ch) {
  if (ch == '\n')
    dos_putc('\r');
  dos_putc(ch);
 }

// Direct console input on DOS

int dos_getchar(void) {
  int ret;

  Asm(	" movb	$6, %%ah	\n\t"	// Direct console I/O
	" movb	$0xFF, %%dl	\n\t"	// Read character
	" int	$0x21		\n\t"	// Execute DOS function call
	" jz	1f		\n\t"	// No data available
	" movb	%%al, %%bl	\n\t"	// BL = AL
	" xorl	%%eax, %%eax	\n\t"	// EAX = 0
	" movb	%%bl, %%al	\n\t"	// EAX = character code
 	" jmp	2f		\n\t"
	"1:			\n\t"
 	" xorl	%%eax, %%eax	\n\t"	// EAX = 0
 	" decl	%%eax		\n\t"	// EAX = -1
	"2:			"
	: "=a" (ret)			// Output register (ret = EAX)
	:				// No input
	: "ebx", "edx"			// Clobbered registers
    );
  return ret;
 }

// Returns current system time on the DOS environment

static
unsigned int dos_readtime(void) {
  unsigned int ret;

  Asm(	" movb	$0x2C, %%ah	\n\t"	// Read system time
	" int	$0x21		\n\t"	// Execute function call
	" xorl	%%eax, %%eax	\n\t"	// Clear EAX
	" movb	%%ch, %%ah	\n\t"	// AH = hours
	" movb	%%cl, %%al	\n\t"	// AL = minutes
	" shll	$16, %%eax	\n\t"	// EAX <<= 16
	" movb	%%dh, %%ah	\n\t"	// AH = seconds
	" movb	%%dl, %%al	"	// AL = centi-seconds
	: "=a" (ret)			// Output register (ret = EAX)
	:				// No input
	: "edx"				// Clobbered register
   );
  return ret;
 }

void readtime(int* hour, int* min, int* sec, int* csec) {
  unsigned int tm;

  tm = dos_readtime();

  *hour = ((tm & 0xFF000000) >> 24) & 0xFF;
  *min  = ((tm & 0x00FF0000) >> 16);
  *sec  = ((tm & 0x0000FF00) >> 8);
  *csec = ((tm & 0x000000FF));
 }

// I/O functions

int inb(int port) {
  int ret;

  Asm(	" xorl	%%eax, %%eax	\n\t"	// EAX = 0
	" inb	%%dx, %%al	"	// Input byte data
	: "=a" (ret)			// Output register (ret = EAX)
	: "d" (port)			// Input register (EDX = port)
   );
  return ret;
 }

int inw(int port) {
  int ret;

  Asm(	" xorl	%%eax, %%eax	\n\t"	// EAX = 0
	" inw	%%dx, %%ax	"	// Input word data
	: "=a" (ret)			// Output register (ret = EAX)
	: "d" (port)			// Input register (EDX = port)
   );
  return ret;
 }

void outb(int port, int data) {
  Asm(	" outb	%%al, %%dx	"	// Output byte data
	:				// No output
	: "a" (data), "d" (port)	// Input registers
					// (EDX = port, EAX = data)
   );
 }

void outw(int port, int data) {
  Asm(	" outw	%%ax, %%dx	"	// Output word data
	:				// No output
	: "a" (data), "d" (port)	// Input registers
					// (EDX = port, EAX = data)
   );
 }

// Tiny standard library function

int atoi(const char* p) {
  int val = 0;

  while (*p) {
    val *= 10;
    val += *p++ - '0';
   }
  return val;
 }
