//
// d u m p v g a . c
//
// Dump VGA registers
//
// Nov. 26, 2003 by Wataru Nishida (http://www.wnishida.com)
//

#include "doslib.h"
#include "printf.h"
#include "vga.h"
#include "realmode.h"

vga_regs v;				// VGA registers holder

void read_crt(void) {
  int i;

  for (i = 0; i < VGA_CRT_REGS; i++) {
    outb(VGA_CRT_ADDR, i);		// Set register index to read
    v.crtc[ i ] = inb(VGA_CRT_DATA);	// Read data
   }
 }

void read_seq(void) {
  int i;

  for (i = 0; i < VGA_SEQ_REGS; i++) {
    outb(VGA_SEQ_ADDR, i);		// Set register index to read
    v.seq[ i ] = inb(VGA_SEQ_DATA);	// Read data
   }
 }

void read_attr(void) {
  int i;

  // Turn off VGA_ATR_LOCK bit while manipulating palette registers!

  for (i = 0; i < VGA_ATR_REGS; i++) {
    inb(VGA_INPUT_STATUS1);		// Flip-flop input status #1 register
    outb(VGA_ATR_ADDWR, i);		// Set register index to read
    v.attr[ i ] = inb(VGA_ATR_READ);	// Read data
   }

  // Turn on the VGA_ATR_LOCK bit again, and recover the screen

  inb(VGA_INPUT_STATUS1);		// Flip-flop input status #1 register
  outb(VGA_ATR_ADDWR, VGA_ATR_LOCK);	// Lock the palette and enable screen
 }

void read_graph(void) {
  int i;

  for (i = 0; i < VGA_GRA_REGS; i++) {
    outb(VGA_GRA_ADDR, i);		// Set register index to read
    v.graph[ i ] = inb(VGA_GRA_DATA);	// Read data
   }
 }

void read_dac(void) {
  int i;

  outb(VGA_DAC_READ, 0);		// Reset register index to zero
  for (i = 0; i < VGA_DAC_REGS; i++) {
    v.dac[ i ] = inb(VGA_DAC_DATA);	// Read data
   }
 }

void read_misc(void) {
  v.misc = inb(VGA_MISC_READ);		// Read miscellaneous output register
 }

void dump_regs(unsigned char* ptr, int len) {
  int i;

  printf("  ");
  for (i = 0; i < len; i++) {
    printf("%02X ", *ptr++);
    if ((i % 16) == 15)
      printf("\n  ");
   }
  if ((i % 16) == 0)
    printf("\n");
  else
    printf("\n\n");
 }

int main(int argc, char** argv) {

  // Read VGA registers

  read_crt();
  read_seq();
  read_attr();
  read_graph();
  read_dac();
  read_misc();

  // Dump VGA registers

  printf("== CRT controller ==\n");
  dump_regs(v.crtc, VGA_CRT_REGS);
  printf("== Sequencer ==\n");
  dump_regs(v.seq, VGA_SEQ_REGS);
  printf("== Attribute controller ==\n");
  dump_regs(v.attr, VGA_ATR_REGS);
  printf("== Graphics controller ==\n");
  dump_regs(v.graph, VGA_GRA_REGS);
  printf("== Miscellaneous output ==\n");
  printf("  %02X\n", v.misc);

  if (argc == 2 && argv[ 1 ][ 0 ] == '-' && argv[ 1 ][ 1 ] == 'd') {
    printf("== DAC controller ==\n");
    dump_regs(v.dac, VGA_DAC_REGS);
   }

  return 0;
 }
