//
// p s f 2 r a w . c
//
// Convert PSF1 font file to raw format
//
// July 5, 2003 by Wataru Nishida (http://www.wnishida.com)
//

#include <stdio.h>	// fprintf(), stderr
#include <stdlib.h>	// exit()
#include <unistd.h>	// read(), write()
#include <fcntl.h>	// open(), O_RDONLY, O_RDWR, O_CREAT, O_TRUNC

// PSF1 header definition

#define PSF1_MAGIC0	0x36	// 1st magic number
#define PSF1_MAGIC1	0x04	// 2nd magic number

#define PSF1_MODE512	0x01	// There are 512 glyphs

typedef struct {
  unsigned char magic[ 2 ];	// Magic number (0x36, 0x04)
  unsigned char mode;		// PSF font mode
  unsigned char height;		// Character height (width is 8 pixels)
 } psf1_header __attribute__ ((packed));

// Error handler

void die(const char* msg) {
  fprintf(stderr, "%s\n", msg);
  exit(1);
 }

int main(int argc, char** argv) {
  int psf, raw, chars, height;
  char buf[ 512 ];
  psf1_header* ph = (psf1_header*) buf;
  int ret, i;

  if (argc != 3)
    die("Usage: psf2raw font.psf font.raw");

  // Open a font file with PSF1 format

  psf = open(argv[ 1 ], O_RDONLY);
  if (psf < 0)
    die("File open failure!");

  // Create a raw format file

  raw = open(argv[ 2 ], O_CREAT | O_TRUNC | O_RDWR, 0666);
  if (raw < 0)
    die("File create failure!");

  // Read header information and check it

  ret = read(psf, buf, sizeof(psf1_header));
  if (ret != sizeof(psf1_header))
    die("Header read error!");

  if (ph->magic[ 0 ] != PSF1_MAGIC0 || ph->magic[ 1 ] != PSF1_MAGIC1)
    die("Illegal PSF1 magic number!");

  chars = (ph->mode & PSF1_MODE512) ? 512 : 256;
  height = ph->height;

  // Output each glyph pattern

  for (i = 0; i < chars; i++) {
    ret = read(psf, buf, height);
    if (ret != height)
      die("Glyph data read error!");
    ret = write(raw, buf, height);
   if (ret != height)
      die("Glyph data write error!");
   }
  
  return 0;
 }
