//
// p s f v i e w . c
//
// PSF type1 font viewer
//
// July 4, 2003 by Wataru Nishida (http://www.wnishida.com)
//

#include <stdio.h>	// fprintf(), stderr
#include <stdlib.h>	// exit()
#include <unistd.h>	// open(), read()
#include <fcntl.h>	// O_RDONLY

// PSF1 header definition

#define PSF1_MAGIC0	0x36	// 1st magic number
#define PSF1_MAGIC1	0x04	// 2nd magic number

typedef struct {
  unsigned char magic[ 2 ];	// Magic number (0x36, 0x04)
  unsigned char mode;		// PSF font mode
  unsigned char height;		// Character height (width is 8 pixels)
 } psf1_header __attribute__ ((packed));

// Error handler

void die(const char* msg) {
  fprintf(stderr, "%s\n", msg);
  exit(1);
 }

int main(int argc, char** argv) {
  int psf, chars = 256, height = -1, fn = 1;
  char buf[ 512 ];
  psf1_header* ph = (psf1_header*) buf;
  int ret, i, j, k;

  // Check arguments

  if (argc < 2 || argc > 3)
    die("Usage: psfview font.psf\n       psfview -xx font.raw");

  if (argc == 3) {
    if (argv[ 1 ][ 0 ] != '-')
      die("Please specify font height! (-8, -16)");
    height = atoi(&argv[ 1 ][ 1 ]);
    if (height < 8 || height >> 16)
      die("Height must be within 8~16!");
    fn = 2;
   }

  // Open a font file

  psf = open(argv[ fn ], O_RDONLY);
  if (psf < 0)
    die("File open failure!");
  if (height != -1)
    goto output;

  // Read header information and check it

  ret = read(psf, buf, sizeof(psf1_header));
  if (ret != sizeof(psf1_header))
    die("Header read error!");

  if (ph->magic[ 0 ] != PSF1_MAGIC0 || ph->magic[ 1 ] != PSF1_MAGIC1)
    die("Illegal PSF1 magic number!");

  height = ph->height;

  // Output each glyph pattern

 output:
  for (i = 0; i < chars; i++) {
    printf("ASCII code %d (0x%02X)\n", i, i);
    ret = read(psf, buf, height);
    if (ret != height)
      die("Glyph data read error!");
    for (j = 0; j < height; j++) {
      for (k = 7; k >= 0; k--) {
        if (buf[ j ] & (1 << k))
          printf("*");
        else
          printf(" ");
       }
      printf("|\n");
     }
    printf("--------+\n");
   }
  
  return 0;
 }
