//
// r a w 2 t x t . c
//
// Convert a raw font data file to text
//
// July 5, 2003 by Wataru Nishida (http://www.wnishida.com)
//

#include <stdio.h>	// fprintf(), stderr
#include <stdlib.h>	// exit()
#include <unistd.h>	// open(), read()
#include <fcntl.h>	// O_RDONLY

// Global variables

int width = 8,			// Character width: 8 bits / line (FIXED)
    height = 0,			// Character height
    chars_line = 8;             // Characters / line

// Error handler

void die(const char* msg) {
  fprintf(stderr, "%s\n", msg);
  exit(1);
 }

int main(int argc, char** argv) {
  int  raw, ret,
       idx, h, i, j, k;
  char buf[ 256 * 16 ];

  // Check arguments

  if (argc != 3)
    die("Usage: raw2txt -xx font.raw");

  if (argv[ 1 ][ 0 ] != '-')
    die("Please specify font height! (-8, -16)");
  height = atoi(&argv[ 1 ][ 1 ]);
  if (height < 8 || height >> 16)
    die("Height must be within 8~16!");

  // Open a font file

  raw = open(argv[ 2 ], O_RDONLY);
  if (raw < 0)
    die("File open failure!");

  // Read font data

  ret = read(raw, (char*) buf, height * 256);
  if (ret != (height * 256))
    die("File read failure!");

  // Output glyph pattern table

  for (h = 0; h < (256 / chars_line); h++) {
    idx = h * chars_line * height;
    for (i = 0; i < height; i++) {
      for (j = 0; j < chars_line; j++) {
        for (k = 7; k >= 0; k--) {
          if (buf[ idx + j * height + i ] & (1 << k))
            printf("*");
          else
            printf(" ");
         }
        printf("|");
       }
      printf("\n");
     }
    for (j = 0; j < chars_line; j++)
      printf("--------+");
    printf("\n");
   }
  
  return 0;
 }
