//
// t e x t . c
//
// Primitive PC/AT VGA text driver
//
// Nov. 25, 2003 by Wataru Nishida (http://www.wnishida.com)
//

#include "vga.h"
#include "text.h"
#include "realmode.h"

// Global scope variable

screen scrn[ TEXT_MAX_SCREENS ] = {	// Screen structure holder
  { (short*) 0x0000, 80, 25, 0, 0, 1, TEXT_DEF_ATTRIBUTE },
  { (short*) 0x1000, 80, 25, 0, 0, 1, TEXT_DEF_ATTRIBUTE },
  { (short*) 0x2000, 80, 25, 0, 0, 1, TEXT_DEF_ATTRIBUTE },
  { (short*) 0x3000, 80, 25, 0, 0, 1, TEXT_DEF_ATTRIBUTE },
  { (short*) 0x4000, 80, 25, 0, 0, 1, TEXT_DEF_ATTRIBUTE },
  { (short*) 0x5000, 80, 25, 0, 0, 1, TEXT_DEF_ATTRIBUTE },
  { (short*) 0x6000, 80, 25, 0, 0, 1, TEXT_DEF_ATTRIBUTE },
  { (short*) 0x7000, 80, 25, 0, 0, 1, TEXT_DEF_ATTRIBUTE }
 };

int scrn_page = 0;			// Current page number

// Functions

void selectpage(int num) {
  if (num < 0)
    scrn_page = 0;
  else if (num >= TEXT_MAX_SCREENS)
    scrn_page = TEXT_MAX_SCREENS - 1;
  else
    scrn_page = num;
 }

void color(int attr) {
  scrn[ scrn_page ].attr = attr;
 }

void locate(int xx, int yy) {
  if (xx < 0)
    scrn[ scrn_page ].x = 0;
  else if (xx >= scrn[ scrn_page ].width)
    scrn[ scrn_page ].x = scrn[ scrn_page ].width - 1;
  else
    scrn[ scrn_page ].x = xx;

  if (yy < 0)
    scrn[ scrn_page ].y = 0;
  else if (yy >= scrn[ scrn_page ].height)
    scrn[ scrn_page ].y = scrn[ scrn_page ].height - 1;
  else
    scrn[ scrn_page ].y = yy;
 }

static void xfer_line(short* src, short* dst, int len) {
  Asm(	" pushw %%es		\n\t"	// Push current ES
	" pushw %%ds		\n\t"	// Push current DS
	" movw $0xB800, %%ax	\n\t"	// DS = ES = text VRAM segment
	" movw %%ax, %%ds	\n\t"
	" movw %%ax, %%es	\n\t"
	" cld			\n\t"	// Clear direction flag
	" rep movsw		\n\t"	// Transfer words
	" popw %%ds		\n\t"	// Pop original DS
	" popw %%es		\n\t"	// Pop original ES
	:				// No output
	: "S" (src),			// ESI = src
	  "D" (dst),			// EDI = dst
	  "c" (len)			// ECX = len
   );
 }

static void fill_line(short* dst, int len, int data) {
  Asm(	" pushw %%es		\n\t"	// Push current ES
	" movw $0xB800, %%ax	\n\t"	// ES = text VRAM segment
	" movw %%ax, %%es	\n\t"
	" movw %%bx, %%ax	\n\t"	// AX = data
	" cld			\n\t"	// Clear direction flag
	" rep stosw		\n\t"	// Store words
	" popw %%es		\n\t"	// Pop original ES
	:				// No output
	: "D" (dst),			// EDI = dst
	  "c" (len),			// ECX = len
	  "b" (data)			// EBX = data
   );
 }

void scrollup(void) {
  int h, w, i;
  short *src, *dst;

  h = scrn[ scrn_page ].height;
  w = scrn[ scrn_page ].width;
  dst = scrn[ scrn_page ].buffer;
  src = dst + w;
  for (i = 1; i < h; i++) {
    xfer_line(src, dst, w);
    src += w; dst += w;
   }
  fill_line(dst, w, vram_data(' ', scrn[ scrn_page ].attr));
 }

void clear(void) {
  int h, w, i;
  short *dst;

  h = scrn[ scrn_page ].height;
  w = scrn[ scrn_page ].width;
  dst = scrn[ scrn_page ].buffer;
  for (i = 0; i < h; i++) {
    fill_line(dst, w, vram_data(' ', scrn[ scrn_page ].attr));
    dst += w;
   }
  locate(0, 0);				// Goto home position
  cursor_pos(0, 0);
 }

static void fill_char(short* dst, int data) {
  Asm(	" pushw %%es		\n\t"	// Push current ES
	" movw $0xB800, %%ax	\n\t"	// ES = text VRAM segment
	" movw %%ax, %%es	\n\t"
	" movw %%bx, %%ax	\n\t"	// AX = data
	" stosw			\n\t"	// Store a word
	" popw %%es		\n\t"	// Pop original ES
	:				// No output
	: "D" (dst),			// EDI = dst
	  "b" (data)			// EBX = data
   );
 }

void putchar(int ch) {
  int h, w;
  short* dst;

  h = scrn[ scrn_page ].height;
  w = scrn[ scrn_page ].width;
  if (ch < ' ') {		// Control code
    switch(ch) {
      case '\n':
        scrn[ scrn_page ].x = 0;
        if (scrn[ scrn_page ].y < (h - 1))
          scrn[ scrn_page ].y++;
        else
          scrollup();
        break;

      default:
        break;
     }
   }
  else {			// Printable character
    ch &= 0xFF;
    dst = scrn[ scrn_page ].buffer;
    dst += scrn[ scrn_page ].y * w + scrn[ scrn_page ].x;
    fill_char(dst,  vram_data(ch, scrn[ scrn_page ].attr));
    if (scrn[ scrn_page ].x < (w - 1))
      scrn[ scrn_page ].x++;
    else {
      scrn[ scrn_page ].x = 0;
      if (scrn[ scrn_page ].y < (h - 1))
        scrn[ scrn_page ].y++;
      else
        scrollup();
     }
   }
  cursor_pos(scrn[ scrn_page ].x, scrn[ scrn_page ].y);
 }
