//
// t x t 2 s r c . c
//
// Convert a text file to font source file
//
// Jul. 12, 2003 by Wataru Nishida (http://www.wnishida.com)
// Dec.  4, 2003 -- Bug fix for 8 pixels font handling
//

#include <stdio.h>	// fopen(), fclose(), fprintf(), stderr
#include <stdlib.h>	// exit()
#include <unistd.h>	// open(), read()
#include <fcntl.h>	// O_RDONLY

// Macro definition

#define MAX_HEIGHT	16	// Maximum height of font

// Global variables

int width = 8,			// Character width: 8 bits / line (FIXED)
    height = 0,			// Character height
    chars_line = 8;		// Characters / line

// Error handler

void die(const char* msg) {
  fprintf(stderr, "%s\n", msg);
  exit(1);
 }

int main(int argc, char** argv) {
  int  txt, ret,
       idx, h, i, j, k, val;
  char line[ 128 ],
       buf[ 256 * MAX_HEIGHT ];

  // Check arguments

  if (argc != 3)
    die("Usage: txt2src -xx font.txt");

  if (argv[ 1 ][ 0 ] != '-')
    die("Please specify font height! (-8, -16)");
  height = atoi(&argv[ 1 ][ 1 ]);
  if (height < 8 || height >> MAX_HEIGHT)
    die("Height must be between 8 and 16!");

  // Open a text file

  txt = open(argv[ 2 ], O_RDONLY);
  if (txt < 0)
    die("File open failure!");

  // Clear buffer

  for (i = 0; i < 256; i++)
    for (j = 0; j < 16; j++)
      buf[ i * 16 + j ] = 0;

  // Read text data and convert it to source data

  for (h = 0; h < (256 / chars_line); h++) {
    idx = h * chars_line * MAX_HEIGHT;
    for (i = 0; i < height; i++) {
      ret = read(txt, line, 9 * chars_line + 1);
      if (ret != (9 * chars_line + 1))
        die("File read failure!");
      for (j = 0; j < chars_line; j++) {
        val = 0;
        for (k = 0; k < 8; k++)
          val += ((line[ j * 9 + k] == '*') ? 1 : 0) << (7 - k);
        buf[ idx + j * MAX_HEIGHT + i ] = val;
       }
     }
    ret = read(txt, line, 9 * chars_line + 1);		// Skip separator
    if (ret != (9 * chars_line + 1))
      die("File read failure!");
   }
  
  // Output the result

  printf("// Font file: %s (8x%d)\n\n", argv[ 2 ], height);
  printf("unsigned char font_data[ 256 ][ 32 ] = {\n");
  for (i = 0; i < 256; i++) {
    printf(" { ");
    for (j = 0; j < 8; j++) {
      printf("0x%02X, ", buf[ i * MAX_HEIGHT + j ] & 255);
     }
    printf("\t\t// ASCII %02X (%d)\n", i, i);
    printf("   ");
    for (j = 8; j < 16; j++) {
      printf("0x%02X, ", buf[ i * MAX_HEIGHT + j ] & 255);
     }
    printf("\n");

    printf("   ");
    for (j = 0; j < 8; j++) {
      printf("0x%02X, ", 0);
     }
    printf("\n");
    printf("   ");
    for (j = 0; j < 8; j++) {
      printf("0x%02X, ", 0);
     }
    printf(" },\n");
   }
  printf(" };\n");

  close(txt);
  return 0;
 }
