//
// v g a . c
//
// VGA module
//
// Dec. 15, 2003 by Wataru Nishida (http://www.wnishida.com)
//

#include "doslib.h"			// outb(), inb()
#include "vga.h"
#include "text.h"
#include "realmode.h"

// File scope variables

static int cursor_start = 13,		// Cursor start line number
           cursor_end   = 14;		// Cursor end line number

// Change cursor size

void cursor_size(int start, int end) {
  Asm("cli");				// Disable interruptions
  outb(VGA_CRT_ADDR, VGA_CRT_CURSTART);	// Select cursor start register
  inb(HARD_WAIT);			// Insert wait
  outb(VGA_CRT_DATA, start);		// Update it
  inb(HARD_WAIT);			// Insert wait
  outb(VGA_CRT_ADDR, VGA_CRT_CUREND);	// Select cursor end register
  inb(HARD_WAIT);			// Insert wait
  outb(VGA_CRT_DATA, end);		// Update it
  Asm("sti");				// Enable interruptions
  cursor_start = start;			// Update line number holders
  cursor_end = end;
 }

#ifndef NO_CURSOR_POS
// Update cursor position

void cursor_pos(int x, int y) {
  int start = cursor_start,		// Remember current line numbers
      end = cursor_end,
      add;				// Cursor address

  add = y * scrn[scrn_page].width;	// Calculate cursor address
  add += x;
  add += (((int) scrn[scrn_page].buffer) / 2);
  cursor_size(1, 0);			// Disable cursor
  Asm("cli");				// Disable interruptions
  outb(VGA_CRT_ADDR, VGA_CRT_CURADDH);	// Select cursor address high register
  inb(HARD_WAIT);			// Insert wait
  outb(VGA_CRT_DATA, add >> 8);		// Set higher address
  inb(HARD_WAIT);			// Insert wait
  outb(VGA_CRT_ADDR, VGA_CRT_CURADDL);	// Select cursor address low register
  inb(HARD_WAIT);			// Insert wait
  outb(VGA_CRT_DATA, add & 255);	// Set lower address
  Asm("sti");				// Enable interruptions
  cursor_size(start, end);		// Enable cursor
 }
#endif

// Setup memory environment

void memory_setup(int oddeven, int map, int rplane, int wplane) {
  int val;

  Asm("cli");				// Disable interruptions

  // Halt the sequencer

  outb(VGA_SEQ_ADDR, VGA_SEQ_RESET);	// Select reset reg
  inb(HARD_WAIT);			// Insert wait
  outb(VGA_SEQ_DATA, VGA_SQ0_AR);	// Turn off SR bit and halt sequencer
  inb(HARD_WAIT);			// Insert wait

  // Update sequencer reigsters

  outb(VGA_SEQ_ADDR, VGA_SEQ_MAPMASK);	// Select map mask reg
  inb(HARD_WAIT);			// Insert wait
  outb(VGA_SEQ_DATA, wplane);		// Load write plane number(s)
  inb(HARD_WAIT);			// Insert wait
  outb(VGA_SEQ_ADDR, VGA_SEQ_MEMMODE);	// Select memory mode reg
  inb(HARD_WAIT);			// Insert wait
  val = oddeven? 0:VGA_SQ4_DISABLE_OEW;	// Set write O/E mode bit
  val |= VGA_SQ4_EXT_MEM;		// Turn on extended memory bit
  outb(VGA_SEQ_DATA, val);		// Load O/E mode in writing
  inb(HARD_WAIT);			// Insert wait

  // Resume the sequencer

  outb(VGA_SEQ_ADDR, VGA_SEQ_RESET);	// Select reset reg
  inb(HARD_WAIT);			// Insert wait
  outb(VGA_SEQ_DATA, VGA_SQ0_SR + VGA_SQ0_AR);	// Turn on SR bit and resume
  inb(HARD_WAIT);			// Insert wait

  // Update graphics reigsters

  outb(VGA_GRA_ADDR, VGA_GRA_READMAP);	// Select read map select register
  inb(HARD_WAIT);			// Insert wait
  if (rplane == MEMPLANE_FONT)		// Set memory plane number
    val = 2;
  else if (rplane == MEMPLANE_ATTR)
    val = 1;
  else
    val = 0;
  outb(VGA_GRA_DATA, val);		// Load read plane number
  inb(HARD_WAIT);			// Insert wait
  outb(VGA_GRA_ADDR, VGA_GRA_GRAPHICS);	// Select graphics mode reg
  inb(HARD_WAIT);			// Insert wait
  val = oddeven? VGA_GC5_ENABLE_OER:0;	// Set read O/E mode bit
  outb(VGA_GRA_DATA, val);		// Load read O/E mode in reading
  inb(HARD_WAIT);			// Insert wait
  outb(VGA_GRA_ADDR, VGA_GRA_MISC);	// Select graphics misc reg
  inb(HARD_WAIT);			// Insert wait
  val = oddeven? VGA_GC6_TURNON_OE : 0;	// Setup O/E mode switch
  val |= map;				// Select memory map
  outb(VGA_GRA_DATA, val);		// Load memory map and O/E switch

  Asm("sti");				// Enable interruptions
 }

// Select memory mode

void memory_mode(int mode) {
  if (mode == MEMORY_TEXT) {
      memory_setup(1, VGA_GC6_MAP3, MEMPLANE_ASCII, MEMPLANE_TEXT); 
   }
  else if (mode == MEMORY_FONT) {
      memory_setup(0, VGA_GC6_MAP1, MEMPLANE_FONT, MEMPLANE_FONT);
   }
 }

// Update display memory start address

void memory_start(int add) {

  // Skip drawing period

  while ((inb(VGA_INPUT_STATUS1) & VGA_IS1_VR) == 0) {
    inb(HARD_WAIT);
   }

  // Skip vertical retrace period

  while ((inb(VGA_INPUT_STATUS1) & VGA_IS1_VR)) {
    inb(HARD_WAIT);
   }

  // Now we can update start address register safely

  Asm("cli");				// Disable interruptions
  outb(VGA_CRT_ADDR, VGA_CRT_STARTADDH);// Select start address high register
  inb(HARD_WAIT);			// Insert wait
  outb(VGA_CRT_DATA, (add>>8) & 0xFF);	// Write upper address
  inb(HARD_WAIT);			// Insert wait
  outb(VGA_CRT_ADDR, VGA_CRT_STARTADDL);// Select start address low register
  inb(HARD_WAIT);			// Insert wait
  outb(VGA_CRT_DATA, add & 0xFF);	// Write lower address
  Asm("sti");				// Enable interruptions
 }

// Transfer local font data to memory plane2

void xfer_font(int segment, unsigned char* buffer, int len) {
  Asm(	" pushw	%%es		\n\t"	// Save current ES
	" movw	%%ax, %%es	\n\t"	// ES  = video memory segment
	" movl	$0, %%edi	\n\t"	// EDI = 0
	" cld			\n\t"	// Clear direction flag
	" rep	movsb		\n\t"	// Transfer data
 	" popw	%%es		"	// Recover ES
	:				// No output
	: "a" (segment),		// EAX = segment
	  "S" (buffer),			// ESI = buffer
	  "c" (len)			// ECX = len
	: "%edi"			// Clobbered registers
   );
 }

// I/O functions for VGA registers

int read_crtreg(int idx) {
  outb(VGA_CRT_ADDR, idx);		// Set register index to read
  return (inb(VGA_CRT_DATA));		// Return data
 }

void write_crtreg(int idx, int val) {
  outb(VGA_CRT_ADDR, idx);		// Set register index to write
  outb(VGA_CRT_DATA, val);		// Write data
 }

int read_seqreg(int idx) {
  outb(VGA_SEQ_ADDR, idx);		// Set register index to read
  return (inb(VGA_SEQ_DATA));		// Return data
 }

void write_seqreg(int idx, int val) {
  outb(VGA_SEQ_ADDR, idx);		// Set register index to write
  outb(VGA_SEQ_DATA, val);		// Write data
 }

int read_miscreg(void) {
  return (inb(VGA_MISC_READ));		// Return data
 }

void write_miscreg(int val) {
  outb(VGA_MISC_WRITE, val);		// Write data
 }

// Horizontal timing adjustment

int htiming(int total, int width, int blank_start, int blank_end, \
      int retrace_start, int retrace_end, int offset, int char_width) {
  int val, msb;

  // Unprotect write access to lower CRTC registers (0-7)

  val = read_crtreg(VGA_CRT_VRETRACE_END);	// Read current value
  val &= ~VGA_CRT_PROTECT;			// Turn off protect bit
  write_crtreg(VGA_CRT_VRETRACE_END, val);	// And write it out

  // Update CRTC registers

  val = total - 5;					// Total size
  write_crtreg(VGA_CRT_HTOTAL, val & 0xFF);
  val = width - 1;					// Screen width
  write_crtreg(VGA_CRT_HWIDTH, val & 0xFF);
  val = blank_start - 1;				// Blanking start
  write_crtreg(VGA_CRT_HBLANK_START, val & 0xFF);
  val = blank_end & 0x1F;				// Extract lower 5 bits
  val |= VGA_CRT_EVRA;					// Turn on EVRA bit
  msb = (blank_end & 0x20) << 2;			// Extract bit5
  write_crtreg(VGA_CRT_HBLANK_END, val);		// Blanking end
  val = retrace_start - 1;				// Retrace start
  write_crtreg(VGA_CRT_HRETRACE_START, val & 0xFF);
  val = retrace_end & 0x1F;				// Extract lower 5 bits
  val |= msb;						// Supplement MSB
  write_crtreg(VGA_CRT_HRETRACE_END, val);		// Retrace start
  write_crtreg(VGA_CRT_HOFFSET, offset / 2);		// Horizontal offset

  // Update sequencer register

  val = read_seqreg(VGA_SEQ_CLOCKMODE);			// Read current value
  if (char_width == 8)					// 8 pixels
    val |= VGA_SQ1_CHARWIDTH;				// Turn on the bit
  else if (char_width == 9)				// 9 pixels
    val &= ~VGA_SQ1_CHARWIDTH;				// Turn off the bit
  else
    return 1;
  write_seqreg(VGA_SEQ_CLOCKMODE, val);			// Update register

  // Protect write access to lower CRTC registers (0-7)

  val = read_crtreg(VGA_CRT_VRETRACE_END);	// Read current value
  val |= VGA_CRT_PROTECT;			// Turn on protect bit
  write_crtreg(VGA_CRT_VRETRACE_END, val);	// And write it out
  return 0;
 }

// Vertical timing adjustment

int vtiming(int total, int height, int blank_start, int blank_end, \
      int retrace_start, int retrace_end, int screen_type, int char_height) {
  int val, ovr, cur, vbs9;

  // Unprotect write access to lower CRTC registers (0-7)

  cur = read_crtreg(VGA_CRT_VRETRACE_END);		// Read current value
  cur &= ~VGA_CRT_PROTECT;				// Turn off protect bit
  write_crtreg(VGA_CRT_VRETRACE_END, cur);		// And write it out

  // Read overflow reigster

  ovr = read_crtreg(VGA_CRT_OVERFLOW);			// Read current value
  ovr &= VGA_CRT_LC8;					// Mask overflow bits

  // Update CRTC registers

  val = total - 2;					// Read total size
  write_crtreg(VGA_CRT_VTOTAL, val & 0xFF);		// Setup lower 8 bits
  ovr += (val & 0x200) ? VGA_CRT_VT9 : 0;		// Extract bit 9
  ovr += (val & 0x100) ? VGA_CRT_VT8 : 0;		// Extract bit 8

  val = height - 1;					// Read screen height
  write_crtreg(VGA_CRT_VDISPLAY_END, val & 0xFF);	// Setup lower 8 bits
  ovr += (val & 0x200) ? VGA_CRT_VDE9 : 0;		// Extract bit 9
  ovr += (val & 0x100) ? VGA_CRT_VDE8 : 0;		// Extract bit 8

  val = blank_start - 1;				// Read blanking start#
  write_crtreg(VGA_CRT_VBLANK_START, val & 0xFF);	// Setup lower 8 bits
  vbs9 = (val & 0x200) ? VGA_CRT_VBS9 : 0;		// Extract bit 9
  ovr += (val & 0x100) ? VGA_CRT_VBS8 : 0;		// Extract bit 8

  val = blank_end;					// Read blanking end#
  write_crtreg(VGA_CRT_VBLANK_END, val & 0xFF);		// Setup lower 8 bits

  val = retrace_start - 1;				// Read retrace start#
  write_crtreg(VGA_CRT_VRETRACE_START, val & 0xFF);	// Setup lower 8 bits
  ovr += (val & 0x200) ? VGA_CRT_VRS9 : 0;		// Extract bit 9
  ovr += (val & 0x100) ? VGA_CRT_VRS8 : 0;		// Extract bit 8

  val = retrace_end;					// Read retrace end#
  cur = read_crtreg(VGA_CRT_VRETRACE_END);		// Read current value
  cur &= ~0x1F;						// Turn off lower 5bits
  cur |= (val & 0x1F);					// Ored with retrace#
  write_crtreg(VGA_CRT_VRETRACE_END, cur);		// Update it

  val = char_height - 1;				// Read max scan line#
  cur = read_crtreg(VGA_CRT_MAXSCAN);			// Read current value
  cur &= ~0x3F;						// Mask VBS9 and line#
  cur |= (val & 0x1F);					// Setup max scan line#
  cur |= vbs9;						// Ored with VBS9
  write_crtreg(VGA_CRT_MAXSCAN, cur);			// And write it out

  write_crtreg(VGA_CRT_OVERFLOW, ovr);			// Update overflow reg

  // Protect write access to lower CRTC registers (0-7)

  val = read_crtreg(VGA_CRT_VRETRACE_END);		// Read current value
  val |= VGA_CRT_PROTECT;				// Turn on protect bit
  write_crtreg(VGA_CRT_VRETRACE_END, val);		// And write it out

  // Update misc register

  cur = read_miscreg();					// Read current value
  cur &= ~0xC0;						// Turn off higher bits
  if (screen_type == 400)				// 400 lines
    cur |= VGA_MISC_400LINES;
  else if (screen_type == 480)				// 480 pixels
    cur |= VGA_MISC_480LINES;
  else
    return 1;
  write_miscreg(cur);					// Update register
  return 0;
 }
