//
// v g a . h
//
// VGA header file
//
// August 22, 2003 by Wataru Nishida (http://www.wnishida.com)
//

#ifndef _VGA_H_
#define _VGA_H_

#define HARD_WAIT		0x80	// Hardware wait

// General constatns

#define VGA_VERT_FREQ		70	// Vertical frequency

#define MEMORY_TEXT		0	// Memory mode for text data
#define MEMORY_FONT		1	// Memory mode for font data

#define MEMPLANE_ASCII		1	// Memory plane #0 for ASCII code
#define MEMPLANE_ATTR		2	// Memory plane #1 for attribute code
#define MEMPLANE_FONT		4	// Memory plane #2 for font data
#define MEMPLANE_TEXT		3	// Memopy planes required for text mode

// CRT controller

#define VGA_CRT_REGS		25	// Number of CRT controller regs
#define VGA_CRT_ADDR		0x3D4	// CRT controller address register
#define VGA_CRT_DATA		0x3D5	// CRT controller data register

#define VGA_CRT_HTOTAL		0	// Horizontal total characters
#define VGA_CRT_HWIDTH		1	// Horizontal screen width
#define VGA_CRT_HBLANK_START	2	// Horizontal blanking start
#define VGA_CRT_HBLANK_END	3	// Horizontal blanking end
#define   VGA_CRT_EVRA		0x80	//   Enable Vertical Retrace Access
#define VGA_CRT_HRETRACE_START	4	// Horizontal retrace start
#define VGA_CRT_HRETRACE_END	5	// Horizontal retrace end
#define VGA_CRT_VTOTAL		6	// Vertical total lines
#define VGA_CRT_OVERFLOW	7	// Mixture of overflowed bits
#define   VGA_CRT_VRS9		0x80	//   Bit 9 of vertical retrace start
#define   VGA_CRT_VDE9		0x40	//   Bit 9 of vertical display end
#define   VGA_CRT_VT9		0x20	//   Bit 9 of vertical total
#define   VGA_CRT_LC8		0x10	//   Bit 8 of line compare
#define   VGA_CRT_VBS8		0x08	//   Bit 8 of vertical blanking start
#define   VGA_CRT_VRS8		0x04	//   Bit 8 of vertical retrace start
#define   VGA_CRT_VDE8		0x02	//   Bit 8 of vertical display end
#define   VGA_CRT_VT8		0x01	//   Bit 8 of vertical total
#define VGA_CRT_MAXSCAN		9	// Maximum scan line# of font
#define   VGA_CRT_LC9		0x40	//   Bit 9 of line compare
#define   VGA_CRT_VBS9		0x20	//   Bit 9 of vertical blanking start
#define VGA_CRT_CURSTART        10      // Cursor start
#define VGA_CRT_CUREND          11      // Cursor end
#define VGA_CRT_STARTADDH	12      // Start address higher byte
#define VGA_CRT_STARTADDL	13      // Start address lower byte
#define VGA_CRT_CURADDH		14	// Cursor address higher byte
#define VGA_CRT_CURADDL		15	// Cursor address lower byte
#define VGA_CRT_VRETRACE_START	16	// Vertical retrace start
#define VGA_CRT_VRETRACE_END	17	// Vertical retrace end
#define   VGA_CRT_PROTECT	0x80	//   Protect bit
#define VGA_CRT_VDISPLAY_END	18	// Vertical display end
#define VGA_CRT_HOFFSET		19	// Horizontal offset size
#define VGA_CRT_VBLANK_START	21	// Vertical blanking start
#define VGA_CRT_VBLANK_END	22	// Vertical blanking end

// Attribute controller

#define VGA_ATR_REGS		21	// Number of attribute controller regs
#define VGA_ATR_ADDWR		0x3C0	// Address / data write register
#define VGA_ATR_READ		0x3C1	// Data read register

#define   VGA_ATR_LOCK		0x20	// Lock the palettes and enable screen
#define   VGA_ATR_OVERSCAN	17	// Overscan color register

// Sequencer

#define VGA_SEQ_REGS		5	// Number of sequencer regs
#define VGA_SEQ_ADDR		0x3C4	// Address register
#define VGA_SEQ_DATA		0x3C5	// Data read/write register

#define VGA_SEQ_RESET		0	//   Reset register
#define VGA_SEQ_CLOCKMODE	1	//   Clocking mode
#define VGA_SEQ_MAPMASK		2	//   Map mask register
#define VGA_SEQ_MEMMODE		4	//   Memory mode register

#define VGA_SQ0_AR		1	// Asynchronous reset (active low)
#define VGA_SQ0_SR		2	// Synchronous reset (active low)

#define VGA_SQ1_CHARWIDTH	1	// Character width (0:9, 1:8 pixels)

#define VGA_SQ4_DISABLE_OEW	0x04	// Disable odd/even mode in writing
#define VGA_SQ4_EXT_MEM		0x02	// Enable extended memory

// Graphics controller

#define VGA_GRA_REGS		9	// Number of graphics controller regs
#define VGA_GRA_ADDR		0x3CE	// Address register
#define VGA_GRA_DATA		0x3CF	// Data read/write register

#define   VGA_GRA_READMAP	4	//   Read map select register
#define   VGA_GRA_GRAPHICS	5	//   Graphics mode register
#define   VGA_GRA_MISC		6	//   Miscellaneous graphics register

#define VGA_GC5_ENABLE_OER	0x10	//   Enable odd/even mode in reading
#define VGA_GC6_TURNON_OE	0x02	//   Turn on odd/even access mode
#define VGA_GC6_MAP0		0x00	//   Memory map A0000-BFFFF
#define VGA_GC6_MAP1		0x04	//   Memory map A0000-AFFFF
#define VGA_GC6_MAP2		0x08	//   Memory map B0000-B7FFF
#define VGA_GC6_MAP3		0x0C	//   Memory map B8000-BFFFF

// DAC controller

#define VGA_DAC_REGS		(256*3)	// Number of DAC color registers
#define VGA_DAC_READ		0x03C7	// Address register for reading
#define VGA_DAC_WRITE		0x03C8	// Address register for writing
#define VGA_DAC_DATA		0x03C9	// Data register

// Miscellaneous registers

#define VGA_MISC_READ		0x3CC	// Miscellaneous output read register
#define VGA_MISC_WRITE		0x3C2	// Miscellaneous output write register
					// Sync pulse polarities
#define   VGA_MISC_400LINES	0x40	//   V sync (+), H sync (-)
#define   VGA_MISC_480LINES	0xC0	//   V sync (-), H sync (-)
#define VGA_INPUT_STATUS1	0x3DA	// Input status #1 register
#define   VGA_IS1_VR		0x08	//   Vertical retrace
#define   VGA_IS1_DD		0x01	//   Display disabled

// VGA registers holder

typedef struct {
  unsigned char  attr[ VGA_ATR_REGS ];	// Attribute controller registers
  unsigned char   seq[ VGA_SEQ_REGS ];	// Sequencer registers
  unsigned char graph[ VGA_GRA_REGS ];	// Graphics controller registers
  unsigned char  crtc[ VGA_CRT_REGS ];	// CRT controller registers
  unsigned char   dac[ VGA_DAC_REGS ];	// DAC color registers
  unsigned char  misc;			// Misclellaneous output register
 } vga_regs __attribute__((packed));

// Prototype definitions

#ifndef NO_CURSOR_POS
void cursor_pos(int x, int y);
#endif

void cursor_size(int start, int end);
void memory_setup(int oddeven, int map, int rplane, int wplane);
void memory_mode(int mode);
void memory_start(int add);
void xfer_font(int segment, unsigned char* buffer, int len);
int  read_crtreg(int idx);
void write_crtreg(int idx, int val);
int  read_seqreg(int idx);
void write_seqreg(int idx, int val);
int  read_miscreg(void);
void write_miscreg(int val);
int  htiming(int total, int width, int blank_start, int blank_end, \
       int retrace_start, int retrace_end, int offset, int char_width);
int  vtiming(int total, int height, int blank_start, int blank_end, \
       int retrace_start, int retrace_end, int screen_type, int char_height);
#endif
